# Copyright (C) 2009 Ferdinand Majerech
# This file is part of IMGCrush
# For conditions of distribution and use, see copyright notice in LICENSE.txt

from shutil import copyfile, move
from util import *
from tempimage import *
from referencefile import *

class FileTester:
    """Tests the generated files and selects which one should be output.

    Removes other temporary files in the process.
    """
    def __init__(self):
        """Constructor.

        Data members:
        Mode       : Mode of file selection. available modes:
                     "q" : choose the smallest file w/ at least given quality.
                     "s" : choose the best file with at most given filesize.
        MinQuality : Minimum quality - used when mode is q . Specified in %,
                     e.g. "10.1%" or in subjective units, e.g. "10.1", that 
                     are closer to JPG quality settings.
        MaxSize    : Maximum filesize - used when mode is s . Specified in %
                     of input filesize, e.g. "10.1%", % of filesize of the
                     reference file, e.g. "10.1r%", kiB, MiB, GiB, e.g. 
                     "10.1kiB", in bytes, e.g. "10342" or in average bits per
                     pixel, e.g. "2.4bpp" .
        FileList   : List of all temporaty images. Contains their filenames,
                     filesizes, quality values and data about how they were
                     saved.
        """
        self.Mode = "q"
        self.MinQuality = "100%"
        self.MaxSize = "50%"
        self.FileList = []
    
    def set_mode(self, m):
        """Sets mode of file selection.

        Arguments:
        m: Mode to set. Valid values are "q" and "s"
        """
        if(m in("q", "s")):
            self.Mode = m
        else:
            warning("wrong value passed by --mode (-m): using default")

    def set_min_quality(self, q):
        """Sets minimum quality used in "q" mode.

        Argumets:
        q: quality to set. Valid values are floats 0-100% or 0-100.
        """
        if q[-1] == "%":
            self.MinQuality = valid_number_str_suffix(q, "100%", "%", 
                                                      "--minquality", 0, 100,
                                                      True)      
        else:
            self.MinQuality = valid_number_str(q, "100%", "--minquality", 0, 
                                               100, True)

    def set_max_size(self, s):
        """Sets maximum size used in "s" mode.

        Arguments:
        s: size to set. Valid values are floats 0-100% , 0-100r% , 
           0-1048576kiB/MiB/GiB , 0-1073741824
        """
        if s.endswith("r%"):
            self.MaxSize = valid_number_str_suffix(s, "50%", "r%", "--maxsize",
                                                   0, 100, True)
        elif s[-1]=="%":
            self.MaxSize = valid_number_str_suffix(s, "50%", "%", "--maxsize",
                                                   0, 100, True)
        elif s.endswith("bpp"):
            self.MaxSize = valid_number_str_suffix(s, "50%", "bpp", "--maxsize"
                                                   , 0, 1024, True)
        elif s.endswith("kiB") or s.endswith("MiB") or s.endswith("GiB"):
            self.MaxSize = valid_number_str_suffix(s, "50%", s[-3 :], 
                                                   "--maxsize", 0, 1048576, 
                                                   True)
        else:
            self.MaxSize = valid_number_str(s, "50%", "--maxsize", 0, 
                                            1073741824, True)
    def abs_min_quality(self):
        """Returns minimum quality in %.

        Parses stored minimum quality data and returns it's equivalent in %.
        """
        if self.MinQuality[-1] == "%":
            return float(self.MinQuality[:-1])
        else:
            #this is... crude.
            return 100.0 * sqrt(sqrt(sqrt(sqrt(sqrt(float(self.MinQuality)/100)))))
    
    def abs_max_size(self, reffile):
        """Returns maximum size in bytes.

        Parses stored maximum size data and returns it's equivalent in bytes.
        
        Arguments:
        reffile : Reference file used for relative max size options (i.e. in %)
        """
        if self.MaxSize.endswith("kiB"):
            return int(float(self.MaxSize[:-3]) * 1024)
        elif self.MaxSize.endswith("MiB"):
            return int(float(self.MaxSize[:-3]) * 1024 * 1024)
        elif self.MaxSize.endswith("GiB"):
            return int(float(self.MaxSize[:-3]) * 1024 * 1024 * 1024)
        elif self.MaxSize.endswith("r%"):
            return int(float(self.MaxSize[:-2]) / 100.0 * reffile.Size)
        elif self.MaxSize[-1]=="%":
            return int(float(self.MaxSize[:-1]) / 100.0 * reffile.InputSize)
        elif self.MaxSize.endswith("bpp"):
            return int(reffile.Width * reffile.Height * 
                       float(self.MaxSize[:-3]) / 8.0)
        else:
            return int(float(self.MaxSize))
                           
    def compare_files_size(self, minqual, bestsize, file, bestfile):
        """Compares size of file and bestfile. Returns True if file is better.

        If files have same filesize, file with better quality is determined
        to be better. If file is better than bestfile, bestfile is deleted.
        """
        fsize = file.Size
        fqual = file.Quality
        bestqual = minqual
        if bestfile is not None:
            bestqual = bestfile.Quality
        #definitely choose smallest size
        if fsize < bestsize and fqual >= minqual:
            if bestfile is not None:
                subprocess.call("rm '" + bestfile.FileName + "'", shell = True)
            return True
        #if size is identical to the best so far, choose the best quality
        elif fsize == bestsize and fqual > bestqual:
            if bestfile is not None:
                subprocess.call("rm '" + bestfile.FileName + "'", shell = True)
            return True
        #useless file, remove it
        else:
            subprocess.call("rm '" + file.FileName + "'", shell = True)
            return False
    
    def compare_files_quality(self, bestqual, maxsize, file, bestfile):
        """Compares quality of file and bestfile.Returns True if file is better.

        If files have same quality, file with better filesize is determined
        to be better. If file is better than bestfile, bestfile is deleted.
        """
        fsize = file.Size
        fqual = file.Quality
        bestsize = maxsize
        if bestfile is not None:
            bestsize = bestfile.Size
        #definitely choose best quality
        if fqual > bestqual and fsize <= maxsize:
            if bestfile is not None:
                subprocess.call("rm '" + bestfile.FileName + "'", shell = True)
            return True
        #if quality is identical to the best so far, choose the best size
        elif fqual == bestqual and fsize < bestsize:
            if bestfile is not None:
                subprocess.call("rm '" + bestfile.FileName + "'", shell = True)
            return True
        #useless file, remove it
        else:
            subprocess.call("rm '" + file.FileName + "'", shell = True)
            return False              

    def select_best_file(self, filelist, reffile, outfilename):
        """Selects which of generated files should be output, deletes others.

        Arguments:
        filelist    : List of generated temporary files to select from.
        reffile     : Data about the current reference file to compare to.
        outfilename : Filename the output will have.
        """
        output("selecting the best file")
        self.FileList = filelist
        bestfile = None
        if self.Mode == "q":
            bestsize = reffile.Size
            minquality = self.abs_min_quality()
            for f in self.FileList:
                if self.compare_files_size(minquality, bestsize, f, bestfile):
                    bestfile = f
                    bestsize = f.Size
        elif self.Mode == "s":
            bestquality = 0.0
            maxsize = self.abs_max_size(reffile)
            for f in self.FileList:
                if self.compare_files_quality(bestquality, maxsize, f, bestfile):
                    bestfile = f
                    bestquality = f.Quality
        if bestfile is not None:
            try:
                move(bestfile.FileName, outfilename + "." + bestfile.FileType)
            except IOError:
                warning("Could not output file. Maybe the user " +
                        "does not have required permissions?")
        else:
            if reffile.Size < reffile.InputSize:
                warning("File " + reffile.InputName + " could not be " +
                        "optimized according to input. Outputting reference " +
                        "file instead.")
                try:
                    copyfile(reffile.Name, outfilename + ".png")
                except IOError:
                    warning("Could not output file. Maybe user " +
                            "does not have required permissions?")
            else:
                warning("File " + reffile.InputName + " could not be "
                        "optimized according to input. Outputting input "
                        "file instead.")
                out = outfilename + \
                      reffile.InputName[reffile.InputName.rfind("."):]
                try:
                    copyfile(reffile.InputName, out)
                except IOError:
                    warning("Could not output file. Maybe user " +
                            "does not have required permissions?")
