# Copyright (C) 2009 Ferdinand Majerech
# This file is part of IMGCrush
# For conditions of distribution and use, see copyright notice in LICENSE.txt

from util import *
from tempimage import *
from imagesettings import *
from referencefile import *

class GIFProcessor:
    """Handles GIF file generation and it's settings.
    """
    
    def __init__(self):
        """Constructor.

        Data members:
        Interlace     : Generate interlaced GIF images?
        BitDepth      : Minimum bit depth to generate images with.
        Optimizations : Level of optimization to use:
                        0 - none
                        1 - lossless optimizations (same results as with none)
                        2 - some optimizations
                        3 - more optimizations
                        4 - all optimizations
        """ 
        self.set_all("light")

    def set_all(self, all):
        """Sets all GIF settings according to given mode.
               
        fastest means all optimizations and least generated files
        extreme means no optimizations and all generated files
        heavy means lossless optimizations only - i.e. same result as extreme
        """
        if all == "extreme":
            self.Interlace = True
            self.Optimizations = 0
            self.BitDepth = 1
        elif all == "heavy":
            self.Interlace = True
            self.Optimizations = 1
            self.BitDepth = 1
        elif all == "medium":
            self.Interlace = False
            self.Optimizations = 2
            self.BitDepth = 3
        elif all == "light":
            self.Interlace = False
            self.Optimizations = 3
            self.BitDepth = 5
        elif all == "fastest":
            self.Interlace = False
            self.Optimizations = 4
            self.BitDepth = 7
        else:
            warning("wrong value passed by --gifall : using default")

    def set_optimizations(self, opt):
        """Set optimization level to use.
        """
        try:
            self.Optimizations = clamp(int(opt),0,5)
        except ValueException:
            warning("wrong value passed by --gifopt : using default")

    def set_interlace(self):
        """Turn generation of interlaced GIF images on.
        """
        self.Interlace = True                 
    
    def set_min_depth(self, d):
        """Sets minimum bit depth to use.
        """
        try:
            if(int(d) in range(1,9)):
                self.BitDepth = int(d)
            else:
                warning("wrong value passed by --g-depth : using default")
        except ValueException:
            warning("wrong value passed by --g-depth : using default")

    def get_interlace_vals(self):
        """Parses stored interlace data, returns list of used interlace values.
        """
        out = [False]
        if self.Interlace:
            out.append(True)
        return out

    def get_bitdepth_vals(self):
        """Parses stored bitdepth data, returns list of used bitdepth values.
        """
        return range(self.BitDepth, 9)

    def get_vals_str(self, str):
        """Returns all possible values of setting requested by input string.
        """
        if str == "interlace":
            return self.get_interlace_vals()
        if str == "bitdepth":
            return self.get_bitdepth_vals()
        else:
            warning("unknown value passed to GIFProcessor.get_vals_str()")

    def get_max_files(self):
        """Returns maximum number of files generated in optimization.

        Used for % feedback.
        """
        if self.Interlace:
            return 2 * len(self.get_bitdepth_vals())
        return len(self.get_bitdepth_vals())

    def process(self, reffile, filenoext, queue=None, proc=0):
        """Generate GIF files according to settings and return a list of them.
        """
        output(str(proc) + " processing gif")
        outfilelist = []
        #number of files generated until now
        filecount = 0.0
        for depth in self.get_bitdepth_vals():
            for inter in self.get_interlace_vals():
                outfilelist.append(TempImage(globals.TEMPDIR + filenoext +
                    str(inter) + str(depth) + ".gif", "gif", 
                    GIFSettings(inter, depth), reffile))
                filecount += 1.0
                output("gif " + str(proc) + " " + str(filecount /
                       self.get_max_files() * 100) + " %", 3)
        if queue is not None:
            queue.put(outfilelist)
        else:
            return outfilelist
