# Copyright (C) 2009 Ferdinand Majerech
# This file is part of IMGCrush
# For conditions of distribution and use, see copyright notice in LICENSE.txt

from tempimage import *
from imagesettings import *
from referencefile import *
class JPGProcessor:
    """Handles JPG file generation and it's settings.
    """
    def __init__(self):
        """Constructor.

        Data members:
        MinQuality    : Minimal JPG quality to use. Can be int 1-100.
        Step          : Step to increase quality by when generating files.
        Interlace     : Generate interlaced JPG images ?
        Optimize      : Generate images with optimized JPG compression?
        SFactors      : Subsampling factors to use when generating images.
                        "standard" means common factors 1x1,2x1 and 2x2 are used.
                        "all" means that all available factors are used.
                        Otherwise this is a string of comma separated factors
                        in axb format.
                        Available factors are:
                        1x1 (no subsampling)
                        1x2
                        1x4
                        2x1
                        2x2
                        2x4
                        4x1
                        4x2
        Grayscale     : Generate grayscale JPG images ?
        Optimizations : Level of optimization to use:
                        0 - none
                        1 - lossless optimizations (same results as with none)
                        2 - some optimizations
                        3 - more optimizations
                        4 - all optimizations
        """
        self.set_all("light")

    def set_all(self, all):
        """Sets all JPG settings according to given mode.
               
        fastest means all optimizations and least generated files
        extreme means no optimizations and all generated files
        heavy means lossless optimizations only - i.e. same result as extreme
        """
        if all == "extreme":
            self.MinQuality = 1
            self.Step = 1
            self.Interlace = True
            self.Optimize = True
            self.SFactors = "all"
            self.Optimizations = 0
            self.Grayscale = True
        elif all == "heavy":
            #cca 50% of execution time of extreme
            self.MinQuality = 1
            self.Step = 1
            self.Interlace = True
            self.Optimize = True
            self.SFactors = "all"
            self.Optimizations = 1
            self.Grayscale = True
        elif all == "medium":
            #cca 20.83% of exectution time of heavy
            self.MinQuality = 1
            self.Step = 3
            self.Interlace = True
            self.Optimize = True
            self.SFactors = "1x1,2x1,2x2,2x4,4x2"
            self.Optimizations = 2
            self.Grayscale = True
        elif all == "light":
            #cca 30% of execution time of medium
            self.MinQuality = 1
            self.Step = 6
            self.Interlace = True
            self.Optimize = True
            self.SFactors = "standard"
            self.Optimizations = 3
            self.Grayscale = False
        elif all == "fastest":
            #cca 22.22% of execution time of light
            self.MinQuality = 1
            self.Step = 9
            self.Interlace = True
            self.Optimize = True
            self.SFactors = "2x2"
            self.Optimizations = 4
            self.Grayscale = False
        else:
            warning("wrong value passed by --jpgall : using default")

    def set_optimizations(self, opt):
        """Set optimization level to use.
        """
        try:
            self.Optimizations = clamp(int(opt),0,5)
        except ValueException:
            warning("wrong value passed by --jpgopt : using default")

    def set_min_quality(self, q):
        """Sets minimum quality to use
        """
        self.MinQuality = valid_number(q, 50, "--jpgqual", 1, 100)
    
    def set_step(self, s):
        """Sets step to increase quality by.
        """
        self.Step = valid_number(s, 50, "--jpgqual", 1, 99)
    
    def set_interlace(self):
        """Turns interlacing on.
        """
        self.Interlace = True

    def set_optimize(self):
        """Turns optimized compression on.
        """
        self.Optimize = True

    def set_sfactors(self, f):
        """Sets subsampling factors to use.
        """
        if f in ("standard", "all"):
            self.SFactors = f
            return
        for fac in f.split(","):
            if fac not in ("1x1", "1x2", "1x4", "2x1", "2x2", "2x4", "4x1",\
                           "4x2"):
                warning("WARNING: wrong value passed by --jpgsampfactors :\
                        using default")
                return
        self.SFactors = f

    def set_grayscale(self, f):
        """Turns generation of grayscale images on.
        """
        self.Grayscale = True

    def get_quality_vals(self):
        """Parses stored quality settings, returns list of used quality values.
        """
        out = [self.MinQuality]
        quality = self.MinQuality
        while quality < 100:
            quality = quality + self.Step
            if quality >100:
                quality = 100
            out.append(quality)
        return out 

    def get_sfactor_vals(self):
        """Parses stored subsampling factor data and returns a list of factors.
        """
        if(self.SFactors == "standard"):
            return ["1x1", "2x1", "2x2"]
        elif(self.SFactors == "all"):
            return ["1x1", "1x2", "1x4", "2x1", "2x2", "2x4", "4x1", "4x2"]
        else:
            return self.SFactors.split(",")
    
    def get_interlace_vals(self):
        """Parses stored interlace data, returns list of used interlace values.
        """
        if self.Interlace:
            return [False, True]
        return [False]

    def get_optimize_vals(self):
        """Parses stored optimize data, returns list of used optimize values.
        """
        if self.Optimize:
            return [False, True]
        return [False]

    def get_grayscale_vals(self):
        """Parses stored grayscale data, returns list of used grayscale values.
        """
        if self.Grayscale:
            return [False, True]
        return [False]

    def get_vals_str(self, str):
        """Returns all possible values of setting requested by input string.
        """
        if str == "quality":
            return self.get_quality_vals()
        elif str == "interlace":
            return self.get_interlace_vals()
        elif str == "optimize":
            return self.get_optimize_vals()
        elif str == "sampfactor":
            return self.get_sfactor_vals()
        elif str == "grayscale":
            return self.get_grayscale_vals()
        else:
            warning("unknown value passed to JPGProcessor.get_vals_str()")
    
    def get_max_files(self):
        """Returns maximum number of files generated in optimization.

        Used for % feedback.
        """
        out = len(self.get_quality_vals()) * len(self.get_optimize_vals()) *\
              len(self.get_interlace_vals()) * len(self.get_sfactor_vals())*\
              len(self.get_grayscale_vals())
        if self.Optimizations > 0:
            out /= 2
        return out

    def process(self, reffile, filenoext, queue=None, proc=0):
        """Generates temporary jpg files and returns a list of them.
        """
        output(str(proc) + " processing jpg")
        outfilelist = []
        #number of files generated until now
        fcount = 0.0
        for quality in self.get_quality_vals():
            for opt in self.get_optimize_vals():
                for inter in self.get_interlace_vals():
                    for gray in self.get_grayscale_vals():
                        #OPTIMIZATION: do not use optimized compression with 
                        #interlace as it does not work. Always use either interlace
                        #or optimized as it is always better than if none is used.
                        if self.Optimizations >= 1:
                            if (opt and inter) or ((not opt) and (not inter)):
                                continue
                        for factor in self.get_sfactor_vals():
                            outfilename = globals.TEMPDIR + filenoext + "_" +factor\
                                          + str(inter) + str(opt) + str(quality)\
                                          + str(gray) + ".jpg";
                            outfilelist.append(TempImage(outfilename, "jpg", 
                                               JPGSettings(quality, inter,
                                               opt, factor, gray), reffile))
                            fcount += 1.0
                            output("jpg " + str(proc) + " " + str(fcount / 
                                   self.get_max_files() * 100)+" %",3)
        if queue is not None:
            queue.put(outfilelist)
            return
        else:
            return outfilelist
